(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 3.0,
MathReader 3.0, or any compatible application. The data for the notebook 
starts with the line of stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[      9147,        304]*)
(*NotebookOutlinePosition[      9785,        327]*)
(*  CellTagsIndexPosition[      9741,        323]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["Case Study: Solving the Quadratic Equation", "Section"],

Cell[TextData[{
  "This notebook is designed to accompany Chapter 6 of Introduction to \
Scientific Programming: Computational Problem Solving Using ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " and C by Joseph L. Zachary.  In it, we will use ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " to learn how to solve the quadratic equation in floating-point arithmetic \
without loss of significance errors.   (AW,  31Jul97)"
}], "Text"],

Cell[CellGroupData[{

Cell["Implementing the Quadratic Formula", "Subsection"],

Cell[TextData[{
  "You are no doubt familiar with the formula for solving quadratic \
equations.  If you've forgotten what it is, ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " can help."
}], "Text"],

Cell[BoxData[
    \(Solve[a*x^2\  + \ b*x\  + \ c\  == \ 0, \ x]\)], "Input"],

Cell["\<\
Let's write a pair of functions that compute these two roots.   One \
will correspond to the first root (containing the + sign), and the other to \
the second root (containing the - sign).\
\>", "Text"],

Cell[BoxData[
    \(rootplus[a_, b_, c_]\  := \ 
      \((\(-b\)\  + \ Sqrt[b^2 - 4*a*c])\)/\((2*a)\)\)], "Input"],

Cell[BoxData[
    \(rootminus[a_, b_, c_]\  := \ 
      \((\(-b\)\  - \ Sqrt[b^2 - 4*a*c])\)/\((2*a)\)\)], "Input"],

Cell["\<\
As a check on these functions, let's try to solve  x^2 + 5x + 6 = \
0.   By factoring the left-hand side, it is easy to see that the two roots \
are -3 and -2.  Let's confirm this with our newly created functions.\
\>", 
  "Text"],

Cell[BoxData[{
    \(xplus\  = \ rootplus[1, 5, 6]\), 
    \(xminus\  = \ rootminus[1, 5, 6]\)}], "Input"],

Cell["\<\
It is always good practice to test functions on simple problems \
with known answers before using them.\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Loss of Significance", "Subsection"],

Cell["\<\
The problem with the quadratic formula is that, when using \
floating-point arithmetic, it may be subject to loss of significance errors \
of the kind discussed in Section 5.3.1 of the book.  We can see this with an \
example.\
\>", "Text"],

Cell["\<\
Suppose we wish to compute both roots of the equation x^2 + 200000x \
- 3 = 0 using floating-point numbers.\
\>", "Text"],

Cell[BoxData[{
    \(xplus\  = \ rootplus[1., \ 200000., \ \(-3.\)]\), 
    \(xminus\  = \ rootminus[1., \ 200000., \ \(-3.\)]\)}], "Input"],

Cell["\<\
One way of checking these answers is to plug them back into the \
equation.  We start with xminus:\
\>", "Text"],

Cell[BoxData[
    \(xminus^2\  + \ 200000*xminus\  - \ 3\)], "Input"],

Cell["Quite good.  Now for xplus:", "Text"],

Cell[BoxData[
    \(xplus^2\  + \ 200000*xplus\  - \ 3\)], "Input"],

Cell["\<\
Whoops.  This is not so good.  It appears that xplus has not been \
computed accurately at all.\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Explanation", "Subsection"],

Cell["\<\
  We have learned that  whereas xminus has been computed  very \
accurately,  xplus was not computed accurately at all.    We lost quite a \
number of digits in the computation of xplus.   Where did the loss of \
significant digits occur?   To find out, let us repeat the  computations \
using rational arithmetic.\
\>", "Text"],

Cell[BoxData[
    \(Simplify[rootplus[1, \ 200000, \ \(-3\)]]\)], "Input"],

Cell[BoxData[
    \(InputForm[N[Sqrt[10000000003]]]\)], "Input"],

Cell[BoxData[
    \(InputForm[\(-100000\)\  + \ N[Sqrt[10000000003]]]\)], "Input"],

Cell["\<\
It should now be clear where the loss of significant digits \
occurred.   The number  Sqrt[10000000003] rounds to 100000.000015.    When \
this gets added to the value -100000, about six digits are cancelled.   The \
places of these digits are taken by meaningless numbers at the end.   This \
leaves us with inadequate  precision in the final answer. \
\>", "Text"],

Cell["\<\
Note that the same cancellation of digits does not take place in \
the computation of  xminus, which we can easily verify.\
\>", "Text"],

Cell[BoxData[
    \(Simplify[rootminus[1, \ 200000, \ \(-3\)]]\)], "Input"],

Cell[BoxData[
    \(InputForm[N[Sqrt[10000000003]]]\)], "Input"],

Cell[BoxData[
    \(InputForm[\(-100000\)\  - \ N[Sqrt[10000000003]]]\)], "Input"],

Cell["\<\
The two numbers that we are combining in the expression above are \
of the same sign, so the leading digits are not cancelled.  If b is negative, \
then xplus may suffer from cancellation error.  If b is positive, then xminus \
may suffer from cancellation error.  This provides a clue to the remedy of \
the problem, which we now discuss.\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Maintaining Significance", "Subsection"],

Cell[TextData[{
  "The question is how to avoid the loss of significance error.   First we do \
a bit of math.  It is easy to show, using a piece of paper or ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ", that the product of xminus and xplus is always c/a:"
}], "Text"],

Cell[BoxData[
    \(Expand[rootplus[a, b, c]\ *\ rootminus[a, b, c]]\)], "Input"],

Cell["\<\
If xminus suffers from cancellation errors,  xplus should be free \
of any such errors.  In that case we could compute  xminus as

                        xminus = (c/a) / xplus.

Likewise, if xplus suffers from cancellation errors, xminus should be ok, in \
which case we compute

                         xplus = (c/a) / xminus.

Let us test this strategy by re-computing the root above which suffered from \
cancellation errors.  (Keep in mind that a = 1, b = 200000, c = -3).\
\>", 
  "Text"],

Cell[BoxData[
    \(xplusnew\  = \ \((\(-3\)/1)\)/rootminus[1., 200000., \(-3.\)]\)], 
  "Input"],

Cell["\<\
Let's test this new root by plugging it back into the equation\
\>",
   "Text"],

Cell[BoxData[
    \(xplusnew^2\  + \ 200000*xplusnew\  - \ 3\)], "Input"],

Cell["This time, the root has been computed to full precision.", "Text"],

Cell[TextData[{
  "Here is another example, in which we solve x^2 - 16340x + 2.  Before doing \
any ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " computations, predict whether it is xplus or xminus that will suffer from \
cancellation errors.  Do this by examining the signs of the constants."
}], "Text"],

Cell[TextData[{
  "Now we can do the ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " computations."
}], "Text"],

Cell[BoxData[
    \(xplus\  = \ rootplus[1., \ \(-16340.\), \ 2.]\)], "Input"],

Cell[BoxData[
    \(xplus^2\  - \ 16340*xplus\  + \ 2\)], "Input"],

Cell[BoxData[
    \(xminus\  = \ rootminus[1., \ \(-16340.\), \ 2.]\)], "Input"],

Cell[BoxData[
    \(xminus^2\  - \ 16340*xminus\  + \ 2\)], "Input"],

Cell[BoxData[
    \(xminusnew\  = \ \((2/1)\)/xplus\)], "Input"],

Cell[BoxData[
    \(xminusnew^2\  - \ 16340*xminusnew\  + \ 2\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Exercises", "Subsection"],

Cell["\<\
 In each case, compute both roots of the equation using the \
standard quadratic formula.  Decide which root suffers from cancellation \
errors by both (a) looking at the signs of the constants and (b) plugging the \
computed roots back into the equation.  Then compute improved values of the \
poorly computed roots by following the technique suggested in this \
notebook.\
\>", "Text"],

Cell["\<\
(a)   2x^2 - 400000x - 1 = 0

(b)   5x^2 + 800000x + 2 = 0
\
\>", "Text"]
}, Closed]]
}, Open  ]]
},
FrontEndVersion->"X 3.0",
ScreenRectangle->{{0, 1280}, {0, 1024}},
WindowSize->{520, 600},
WindowMargins->{{221, Automatic}, {Automatic, 120}}
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1731, 51, 61, 0, 54, "Section"],
Cell[1795, 53, 457, 10, 86, "Text"],

Cell[CellGroupData[{
Cell[2277, 67, 56, 0, 45, "Subsection"],
Cell[2336, 69, 211, 6, 50, "Text"],
Cell[2550, 77, 77, 1, 27, "Input"],
Cell[2630, 80, 212, 4, 50, "Text"],
Cell[2845, 86, 114, 2, 27, "Input"],
Cell[2962, 90, 115, 2, 27, "Input"],
Cell[3080, 94, 240, 5, 68, "Text"],
Cell[3323, 101, 106, 2, 43, "Input"],
Cell[3432, 105, 127, 3, 50, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[3596, 113, 42, 0, 29, "Subsection"],
Cell[3641, 115, 250, 5, 68, "Text"],
Cell[3894, 122, 131, 3, 50, "Text"],
Cell[4028, 127, 140, 2, 43, "Input"],
Cell[4171, 131, 122, 3, 50, "Text"],
Cell[4296, 136, 69, 1, 27, "Input"],
Cell[4368, 139, 43, 0, 32, "Text"],
Cell[4414, 141, 67, 1, 27, "Input"],
Cell[4484, 144, 119, 3, 32, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[4640, 152, 33, 0, 29, "Subsection"],
Cell[4676, 154, 338, 6, 86, "Text"],
Cell[5017, 162, 74, 1, 27, "Input"],
Cell[5094, 165, 64, 1, 27, "Input"],
Cell[5161, 168, 82, 1, 27, "Input"],
Cell[5246, 171, 376, 6, 86, "Text"],
Cell[5625, 179, 146, 3, 50, "Text"],
Cell[5774, 184, 75, 1, 27, "Input"],
Cell[5852, 187, 64, 1, 27, "Input"],
Cell[5919, 190, 82, 1, 27, "Input"],
Cell[6004, 193, 363, 6, 86, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[6404, 204, 46, 0, 29, "Subsection"],
Cell[6453, 206, 282, 6, 68, "Text"],
Cell[6738, 214, 81, 1, 27, "Input"],
Cell[6822, 217, 506, 14, 230, "Text"],
Cell[7331, 233, 97, 2, 27, "Input"],
Cell[7431, 237, 89, 3, 32, "Text"],
Cell[7523, 242, 73, 1, 27, "Input"],
Cell[7599, 245, 72, 0, 32, "Text"],
Cell[7674, 247, 318, 7, 68, "Text"],
Cell[7995, 256, 122, 5, 32, "Text"],
Cell[8120, 263, 78, 1, 27, "Input"],
Cell[8201, 266, 66, 1, 27, "Input"],
Cell[8270, 269, 80, 1, 27, "Input"],
Cell[8353, 272, 68, 1, 27, "Input"],
Cell[8424, 275, 64, 1, 27, "Input"],
Cell[8491, 278, 74, 1, 27, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[8602, 284, 31, 0, 29, "Subsection"],
Cell[8636, 286, 397, 7, 104, "Text"],
Cell[9036, 295, 83, 5, 86, "Text"]
}, Closed]]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

